package edu.stanford.nlp.parser.lexparser;

import java.util.*;

/**
 * A HookChart is a chart data structure designed for use with the efficient
 * O(n^4) chart parsing mechanisms targetted at lexicalized parsing, which
 * were introduced by Eisner and Satta.
 *
 * @author Dan Klein
 */
class HookChart {

  private Map registeredPreHooks = new HashMap();
  private Map registeredPostHooks = new HashMap();
  private Map registeredEdgesByLeftIndex = new HashMap();
  private Map registeredEdgesByRightIndex = new HashMap();

  private Map<WeakChartIndex,Collection> realEdgesByL = new HashMap<WeakChartIndex,Collection>();
  private Map<WeakChartIndex,Collection> realEdgesByR = new HashMap<WeakChartIndex,Collection>();
  private Set<ChartIndex> builtLIndexes = new HashSet<ChartIndex>();
  private Set<ChartIndex> builtRIndexes = new HashSet<ChartIndex>();

  private Interner interner = new Interner();

  private static class ChartIndex {
    public int state;
    public int head;
    public int tag;
    public int loc;

    public int hashCode() {
      return state ^ (head << 8) ^ (tag << 16) ^ (loc << 24);
    }

    public boolean equals(Object o) {
      if (this == o) {
        return true;
      }
      if (o instanceof ChartIndex) {
        ChartIndex ci = (ChartIndex) o;
        return state == ci.state && head == ci.head && tag == ci.tag && loc == ci.loc;
      }
      return false;
    }

  } // end class ChartIndex

  private static class WeakChartIndex {
    public int state;
    public int loc;

    public int hashCode() {
      return state ^ (loc << 16);
    }

    public boolean equals(Object o) {
      if (this == o) {
        return true;
      }
      if (o instanceof WeakChartIndex) {
        WeakChartIndex ci = (WeakChartIndex) o;
        return state == ci.state && loc == ci.loc;
      }
      return false;
    }
  }

  private static final Collection empty = new ArrayList(0);
  private WeakChartIndex tempWeakIndex = new WeakChartIndex();

  public void registerEdgeIndexes(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.start;
    ChartIndex index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    builtLIndexes.add(index);
    //System.out.println("Edge registered: "+edge);
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.end;
    index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    builtRIndexes.add(index);
  }

  public void registerRealEdge(Edge edge) {
    tempWeakIndex.state = edge.state;
    tempWeakIndex.loc = edge.start;
    WeakChartIndex index = (WeakChartIndex) interner.intern(tempWeakIndex);
    if (index == tempWeakIndex) {
      tempWeakIndex = new WeakChartIndex();
    }
    insert(realEdgesByL, index, edge);
    tempWeakIndex.state = edge.state;
    tempWeakIndex.loc = edge.end;
    index = (WeakChartIndex) interner.intern(tempWeakIndex);
    if (index == tempWeakIndex) {
      tempWeakIndex = new WeakChartIndex();
    }
    insert(realEdgesByR, index, edge);
  }

  public boolean isBuiltL(int state, int start, int head, int tag) {
    tempIndex.state = state;
    tempIndex.head = head;
    tempIndex.tag = tag;
    tempIndex.loc = start;
    return builtLIndexes.contains(tempIndex);
  }

  public boolean isBuiltR(int state, int end, int head, int tag) {
    tempIndex.state = state;
    tempIndex.head = head;
    tempIndex.tag = tag;
    tempIndex.loc = end;
    return builtRIndexes.contains(tempIndex);
  }

  public Collection getRealEdgesWithL(int state, int start) {
    tempWeakIndex.state = state;
    tempWeakIndex.loc = start;
    Collection edges = realEdgesByL.get(tempWeakIndex);
    if (edges == null) {
      return empty;
    }
    return edges;
  }

  public Collection getRealEdgesWithR(int state, int end) {
    tempWeakIndex.state = state;
    tempWeakIndex.loc = end;
    Collection edges = realEdgesByR.get(tempWeakIndex);
    if (edges == null) {
      return empty;
    }
    return edges;
  }

  public Collection getPreHooks(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.end;
    Collection result = (Collection) registeredPreHooks.get(tempIndex);
    if (result == null) {
      result = empty;
    }
    //System.out.println("For "+edge+" returning "+result.size()+" pre hooks");
    return result;
  }

  public Collection getPostHooks(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    tempIndex.loc = edge.start;
    Collection result = (Collection) registeredPostHooks.get(tempIndex);
    if (result == null) {
      result = empty;
    }
    //System.out.println("For "+edge+" returning "+result.size()+" post hooks");
    return result;
  }

  public Collection getEdges(Hook hook) {
    tempIndex.state = hook.subState;
    tempIndex.head = hook.head;
    tempIndex.tag = hook.tag;
    Collection result;
    if (hook.isPreHook()) {
      tempIndex.loc = hook.start;
      result = (Collection) registeredEdgesByRightIndex.get(tempIndex);
    } else {
      tempIndex.loc = hook.end;
      result = (Collection) registeredEdgesByLeftIndex.get(tempIndex);
    }
    if (result == null) {
      result = empty;
    }
    //System.out.println("For "+hook+" returning "+result.size()+" edges");
    return result;
  }

  // TODO: This hacks up a CollectionValuedMap.  Convert to use that class.
  private static void insert(Map map, Object index, Object item) {
    List list = (List) map.get(index);
    if (list == null) {
      list = new ArrayList();
      map.put(index, list);
    }
    list.add(item);
  }

  ChartIndex tempIndex = new ChartIndex();

  public void addEdge(Edge edge) {
    tempIndex.state = edge.state;
    tempIndex.head = edge.head;
    tempIndex.tag = edge.tag;
    // left index
    tempIndex.loc = edge.start;
    ChartIndex index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
      tempIndex.state = edge.state;
      tempIndex.head = edge.head;
      tempIndex.tag = edge.tag;
    }
    insert(registeredEdgesByLeftIndex, index, edge);
    tempIndex.loc = edge.end;
    index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    insert(registeredEdgesByRightIndex, index, edge);
  }

  public void addHook(Hook hook) {
    Map map;
    tempIndex.state = hook.subState;
    tempIndex.head = hook.head;
    tempIndex.tag = hook.tag;
    if (hook.isPreHook()) {
      tempIndex.loc = hook.start;
      map = registeredPreHooks;
    } else {
      tempIndex.loc = hook.end;
      map = registeredPostHooks;
    }
    ChartIndex index = (ChartIndex) interner.intern(tempIndex);
    if (index == tempIndex) {
      tempIndex = new ChartIndex();
    }
    insert(map, index, hook);
  }
} // end class HookChart
