package edu.stanford.nlp.parser.lexparser;

import edu.stanford.nlp.util.Numberer;

import java.io.*;
import java.util.*;

/**
 * Maintains efficient indexing of unary grammar rules.
 *
 * @author Dan Klein
 * @author Christopher Manning
 */
public class UnaryGrammar implements Serializable, Iterable<UnaryRule> {

  /** The number of UnaryRule items in the UnaryGrammar. */
  private int numStates = -1;
  /** The name of the Numberer state space for unary rules. */
  private String stateSpace;

  private transient List<UnaryRule>[] rulesWithParent = null;
  private transient List<UnaryRule>[] rulesWithChild = null;

  private transient List<UnaryRule>[] closedRulesWithParent = null;
  private transient List<UnaryRule>[] closedRulesWithChild = null;

  private transient UnaryRule[][] closedRulesWithP = null;
  private transient UnaryRule[][] closedRulesWithC = null;

  // cdm 2006: why is this needed as Map?  Why not array or List??
  private Map<UnaryRule,UnaryRule> coreRules = null;
  private transient Map<UnaryRule,UnaryRule> bestRulesUnderMax = null;
  private transient Map<UnaryRule,Integer> backTrace = null;

  public int numClosedRules() {
    return bestRulesUnderMax.keySet().size();
  }

  public Iterator<UnaryRule> closedRuleIterator() {
    return bestRulesUnderMax.keySet().iterator();
  }

  public int numRules() {
    return coreRules.keySet().size();
  }

  public Iterator<UnaryRule> iterator() {
    return ruleIterator();
  }

  public Iterator<UnaryRule> ruleIterator() {
    return coreRules.keySet().iterator();
  }

  public List<UnaryRule> rules() {
    return new ArrayList<UnaryRule>(coreRules.keySet());
  }

  /** Remove A -&gt; A UnaryRules from bestRulesUnderMax. */
  public void purgeRules() {
    Map<UnaryRule,UnaryRule> bR = new HashMap<UnaryRule,UnaryRule>();
    for (UnaryRule ur : bestRulesUnderMax.keySet()) {
      if (ur.parent != ur.child) {
        bR.put(ur, ur);
      }
    }
    bestRulesUnderMax = bR;
  }

  public List<Integer> getBestPath(int parent, int child) {
    List<Integer> path = new ArrayList<Integer>();
    UnaryRule tempR = new UnaryRule();
    tempR.parent = parent;
    tempR.child = child;
    //System.out.println("Building path...");
    int loc = parent;
    while (loc != child) {
      path.add(new Integer(loc));
      //System.out.println("Path is "+path);
      tempR.parent = loc;
      Integer nextInt = backTrace.get(tempR);
      if (nextInt == null) {
        loc = child;
      } else {
        loc = nextInt.intValue();
      }
      //System.out.println(Numberer.getGlobalNumberer("states").object(parent)+"->"+Numberer.getGlobalNumberer("states").object(child)+" went via "+Numberer.getGlobalNumberer("states").object(loc));
      if (path.size() > 10) {
        throw new RuntimeException("UnaryGrammar path > 10");
      }
    }
    path.add(new Integer(child));
    return path;
  }

  private void closeRulesUnderMax(UnaryRule ur) {
    for (int i = 0; i < closedRulesWithChild[ur.parent].size(); i++) {
      UnaryRule pr = closedRulesWithChild[ur.parent].get(i);
      for (int j = 0; j < closedRulesWithParent[ur.child].size(); j++) {
        UnaryRule cr = closedRulesWithParent[ur.child].get(j);
        UnaryRule resultR = new UnaryRule();
        resultR.parent = pr.parent;
        resultR.child = cr.child;
        resultR.score = pr.score + cr.score + ur.score;
        if (relaxRule(resultR)) {
          if (resultR.parent != ur.parent) {
            backTrace.put(resultR, new Integer(ur.parent));
          } else {
            backTrace.put(resultR, new Integer(ur.child));
          }
        }
      }
    }
  }

  private boolean relaxRule(UnaryRule ur) {
    UnaryRule bestR = bestRulesUnderMax.get(ur);
    if (bestR == null) {
      bestRulesUnderMax.put(ur, ur);
      closedRulesWithParent[ur.parent].add(ur);
      closedRulesWithChild[ur.child].add(ur);
      return true;
    } else {
      if (bestR.score < ur.score) {
        bestR.score = ur.score;
        return true;
      }
      return false;
    }
  }

  public double scoreRule(UnaryRule ur) {
    UnaryRule bestR = bestRulesUnderMax.get(ur);
    return (bestR != null ? bestR.score : Double.NEGATIVE_INFINITY);
  }

  public void addRule(UnaryRule ur) {
    // add rules' closure
    closeRulesUnderMax(ur);
    coreRules.put(ur, ur);
    rulesWithParent[ur.parent].add(ur);
    rulesWithChild[ur.child].add(ur);
  }

  //public Iterator closedRuleIterator() {
  //  return bestRulesUnderMax.keySet().iterator();
  //}

  private static final UnaryRule[] EMPTY_UNARY_RULE_ARRAY = new UnaryRule[0];

  void makeCRArrays() {

    // this just shadowed class variable: int numStates = closedRulesWithParent.length;
    closedRulesWithP = new UnaryRule[numStates][];
    closedRulesWithC = new UnaryRule[numStates][];
    for (int i = 0; i < numStates; i++) {
      closedRulesWithP[i] = closedRulesWithParent[i].toArray(EMPTY_UNARY_RULE_ARRAY);
      closedRulesWithC[i] = closedRulesWithChild[i].toArray(EMPTY_UNARY_RULE_ARRAY);
    }
  }

  public UnaryRule[] closedRulesByParent(int state) {
    if (closedRulesWithP == null) {
      makeCRArrays();
    }
    if (state >= closedRulesWithP.length) {
      return EMPTY_UNARY_RULE_ARRAY;
    }
    return closedRulesWithP[state];
  }

  public UnaryRule[] closedRulesByChild(int state) {
    if (closedRulesWithC == null) {
      makeCRArrays();
    }
    if (state >= closedRulesWithC.length) {
      return EMPTY_UNARY_RULE_ARRAY;
    }
    return closedRulesWithC[state];
  }

  public Iterator<UnaryRule> closedRuleIteratorByParent(int state) {
    if (state >= closedRulesWithParent.length) {
      List<UnaryRule> lur = Collections.emptyList();
      return lur.iterator();
    }
    return closedRulesWithParent[state].iterator();
  }

  public Iterator<UnaryRule> closedRuleIteratorByChild(int state) {
    if (state >= closedRulesWithChild.length) {
      List<UnaryRule> lur = Collections.emptyList();
      return lur.iterator();
    }
    return closedRulesWithChild[state].iterator();
  }

  public Iterator<UnaryRule> ruleIteratorByParent(int state) {
    if (state >= rulesWithParent.length) {
      List<UnaryRule> lur = Collections.emptyList();
      return lur.iterator();
    }
    return rulesWithParent[state].iterator();
  }

  public Iterator<UnaryRule> ruleIteratorByChild(int state) {
    if (state >= rulesWithChild.length) {
      List<UnaryRule> lur = Collections.emptyList();
      return lur.iterator();
    }
    return rulesWithChild[state].iterator();
  }

  public List<UnaryRule> rulesByParent(int state) {
    if (state >= rulesWithParent.length) {
      List<UnaryRule> lur = Collections.emptyList();
      return lur;
    }
    return rulesWithParent[state];
  }

  public List<UnaryRule> rulesByChild(int state) {
    if (state >= rulesWithChild.length) {
      List<UnaryRule> lur = Collections.emptyList();
      return lur;
    }
    return rulesWithChild[state];
  }

  public List[] rulesWithParent() {
    return rulesWithParent;
  }

  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject();
    Set<UnaryRule> allRules = new HashSet<UnaryRule>(coreRules.keySet());
    init();
    for (UnaryRule ur : allRules) {
      addRule(ur);
    }
    purgeRules();
  }

  /** Create all the array variables,andput in A -&gt; A UnaryRules to feed
   *  the closure algorithm. They then get deleted later.
   */
  private void init() {
    coreRules = new HashMap<UnaryRule,UnaryRule>();
    rulesWithParent = new List[numStates];
    rulesWithChild = new List[numStates];
    closedRulesWithParent = new List[numStates];
    closedRulesWithChild = new List[numStates];
    bestRulesUnderMax = new HashMap<UnaryRule,UnaryRule>();
    backTrace = new HashMap<UnaryRule,Integer>();
    for (int s = 0; s < numStates; s++) {
      rulesWithParent[s] = new ArrayList<UnaryRule>();
      rulesWithChild[s] = new ArrayList<UnaryRule>();
      closedRulesWithParent[s] = new ArrayList<UnaryRule>();
      closedRulesWithChild[s] = new ArrayList<UnaryRule>();
      UnaryRule selfR = new UnaryRule(s, s, 0.0);
      relaxRule(selfR);
    }
  }

  public UnaryGrammar(int numStates) {
    this(numStates, "states");
  }

  public UnaryGrammar(int numStates, String stateSpace) {
    this.numStates = numStates;
    this.stateSpace = stateSpace;
    init();
  }

  /**
   * Populates data in this UnaryGrammar from the character stream
   * given by the Reader in.
   */
  public void readData(BufferedReader in) throws IOException {
    String line;
    int lineNum = 1;
    Numberer n = Numberer.getGlobalNumberer(stateSpace);
    // all lines have one rule per line
    line = in.readLine();
    while (line != null && line.length() > 0) {
      try {
        addRule(new UnaryRule(line, n));
      } catch (Exception e) {
        throw new IOException("Error on line " + lineNum);
      }
      lineNum++;
      line = in.readLine();
    }
    purgeRules();
  }

  /**
   * Writes out data from this Object to the Writer w.
   */
  public void writeData(Writer w) throws IOException {
    PrintWriter out = new PrintWriter(w);
    // all lines have one rule per line
    for (Iterator rI = ruleIterator(); rI.hasNext();) {
      out.println(rI.next().toString());
    }
    out.flush();
  }

  public String toString() {
    Writer w = new StringWriter();
    try {
      writeData(w);
    } catch (IOException ioe) {
    }
    return w.toString();
  }

  private static final long serialVersionUID = 1L;

} // end class UnaryGrammar
