package edu.stanford.nlp.process;

import edu.stanford.nlp.ling.StringLabel;
import edu.stanford.nlp.ling.Word;
import edu.stanford.nlp.ling.WordLemmaTag;
import edu.stanford.nlp.ling.WordTag;
import edu.stanford.nlp.util.StringUtils;

import java.io.FileReader;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;


/**
 * Morphology computes the base form of English words, by removing just
 * inflections (not derivational morphology).  It is based on a finite-state
 * transducer implemented by John Carroll et al., written in flex and publicly
 * available.
 * See: http://www.informatics.susx.ac.uk/research/nlp/carroll/morph.html .
 * There are several ways of invoking Morphology. One is by calling the static
 * methods
 * WordTag stemStatic(String word, String tag) or
 * WordTag stemStatic(WordTag wordTag).
 * If we have created a Morphology object already we can use the methods
 * WordTag stem(String word, string tag) or WordTag stem(WordTag wordTag).
 * <p/>
 * Another way of using Morphology is to run it on an input file by running
 * <code>java Morphology filename</code>
 *
 * @author Kristina Toutanova (kristina@cs.stanford.edu)
 * @author Christopher Manning
 */
public class Morphology implements Function {

  private static final boolean DEBUG = false;

  private Morpha lexer;
  private static Morpha staticLexer;

  public Morphology() {
    lexer = new Morpha(System.in);
  }

  /**
   * Process morphologically words from a Reader.
   */
  public Morphology(Reader in) {
    lexer = new Morpha(in);
  }

  public Morphology(String filename) {
    try {
      lexer = new Morpha(new FileReader(filename));
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  public Word next() throws IOException {
    return new Word(lexer.next());
  }

  static boolean isProper(String posTag) {
    return posTag.equals("NNP") || posTag.equals("NNPS") || posTag.equals("NP");
  }

  public Word stem(Word w) {
    try {
      lexer.yyreset(new StringReader(w.value()));
      lexer.yybegin(Morpha.any);
      StringLabel wordRes = lexer.next();
      return new Word(wordRes.value());
    } catch (Exception e) {
      e.printStackTrace();
    }
    return w;
  }

  public WordTag stem(WordTag wT) {
    return stem(wT.word(), wT.tag());
  }

  public WordTag stem(String word, String tag) {
    return stem(word, tag, lexer, lexer.option(1));
  }

  /** Lemmatize the word, being sensitive to the tag, using the
   *  passed in lexer.
   *
   *  @param lowercase If this is true, words other than proper nouns will
   *      be changed to all lowercase.
   */
  public static WordTag stem(String word, String tag, Morpha lexer,
                             boolean lowercase) {
    boolean wordHasForbiddenChar = word.indexOf("_") >= 0 || 
      word.indexOf(" ") >= 0;
    String quotedWord = word;
    if (wordHasForbiddenChar) {
      try {
        // choose something unlikely. Devangari!
        quotedWord = quotedWord.replaceAll("_", "\u0960");
        quotedWord = quotedWord.replaceAll(" ", "\u0961");
      } catch (Exception e) {
        System.err.println("stem: Didn't work");
      }
    }
    String wordtag = quotedWord + "_" + tag;
    if (DEBUG) System.err.println("Trying to normalize |" + wordtag + "|");
    try {
      lexer.setOption(1, lowercase);
      lexer.yyreset(new StringReader(wordtag));
      lexer.yybegin(Morpha.scan);
      StringLabel wordRes = lexer.next();
      StringLabel tagRes = lexer.next();
      if (wordHasForbiddenChar) {
        try {
          if (DEBUG) System.err.println("Restoring forbidden chars");
          String w = wordRes.value();
          w = w.replaceAll("\u0960", "_");
          w = w.replaceAll("\u0961", " ");
          wordRes.setValue(w);
        } catch (Exception e) {
          System.err.println("stem: Didn't work");
        }
      }
      return new WordTag(wordRes.value(), tag);
    } catch (Throwable e) {
      System.err.println("Morphology.stem() had error on word " + word + "/" + tag);
      return new WordTag(word, tag);
    }
  }


  public static WordTag stemStatic(String word, String tag) {
    if (staticLexer == null) {
      staticLexer = new Morpha(System.in);
    }
    return stem(word, tag, staticLexer, staticLexer.option(1));
  }

  public static WordTag stemStatic(String word, String tag,
                                   boolean lowercase) {
    if (staticLexer == null) {
      staticLexer = new Morpha(System.in);
    }
    return stem(word, tag, staticLexer, lowercase);
  }

  /** Return a new WordTag which has the lemma as the value of word().
   *  The default is to lowercase non-proper-nouns, unless options have
   *  been set.
   */
  public static WordTag stemStatic(WordTag wT) {
    return stemStatic(wT.word(), wT.tag());
  }


  public Object apply(Object in) {
    if (in instanceof WordTag) {
      return stem((WordTag) in);
    }
    if (in instanceof Word) {
      return stem((Word) in);
    }
    return in;
  }

  /**
   * Lemmatize returning a <code>WordLemmaTag </code>.
   */
  public WordLemmaTag lemmatize(WordTag wT) {
    String tag = wT.tag();
    String word = wT.word();
    String lemma = stem(wT).word();
    return new WordLemmaTag(word, lemma, tag);
  }

  public static WordLemmaTag lemmatizeStatic(WordTag wT) {
    String tag = wT.tag();
    String word = wT.word();
    String lemma = stemStatic(wT).word();
    return new WordLemmaTag(word, lemma, tag);
  }


  public static void main(String[] args) throws IOException {
    if (args.length == 0) {
      System.out.println("John/NNP -> " +
          stemStatic(new WordTag("John", "NNP")));
      System.out.println("Corporations/NNPS -> " +
          stemStatic(new WordTag("Corporations", "NNPS")));
      WordTag hunt = new WordTag("hunting", "V");
      System.out.println("hunting/V -> " + stemStatic(hunt));
      System.out.println("hunting/V -> " + lemmatizeStatic(hunt));
      String[] exWords = new String[]{"brethren", "ducks", "saw", "saw",
                                      "running", "runing", "making", "makking",
                                      "stopped", "stoped", "xopped", "won't",
                                      "cleaner", "cleaner", "took", "bought",
                                      "am", "were", "did", "n't", "wo",
                                      "'s", "'s", "ca", "her", "her", "their",
                                      "Books", "light-weight", "cease-fire",
                                      "John_William_Smith", "Dogs" };
      String[] exTags = new String[]{"NNS", "NNS", "VBD", "NN",
                                     "VBG", "VBG", "VBG", "VBG",
                                     "VBD", "VBD", "VBD", "MD",
                                     "NN", "JJR", "VBD", "VBD",
                                     "VBP", "VBD", "VBD", "RB", "MD",
                                     "VBZ", "POS", "MD", "PRP", "PRP$", "PRP$",
                                     "NNPS", "JJ", "NN",
                                     "NNP", "NNS" };
      for (int i = 0; i < exWords.length; i++) {
        System.out.println(exWords[i] + "/" + exTags[i] + " -> " +
            lemmatizeStatic(new WordTag(exWords[i], exTags[i])));
      }
    } else if (args.length == 2 && args[0].equals("-rebuildVerbTable")) {
      String verbs = StringUtils.slurpFile(args[1]);
      String[] words = verbs.split("\\s+");
      System.out.print(" private static String[] verbStems = new String[] { ");
      for (int i = 0; i < words.length; i++) {
        System.out.print("\"" + words[i] + "\"");
        if (i != words.length - 1) {
          System.out.print(", ");
          if (i % 5 == 0) {
            System.out.println();
            System.out.print("    ");
          }
        }
      }
      System.out.println(" };");
    } else if (args[0].equals("-stem")) {
      for (int i = 1; i < args.length; i++) {
        System.out.println(args[i] + " --> " + stemStatic(WordTag.valueOf(args[i])));
      }
    } else {
      for (String arg :  args) {
        Morphology morph = new Morphology(arg);
        Word next;
        while ((next = morph.next()) != null) {
          System.out.print(next);
        }
      }
    }
  }

  private static final long serialVersionUID = 2;

}
