package edu.stanford.nlp.trees;

import edu.stanford.nlp.ling.Label;
import edu.stanford.nlp.parser.lexparser.LexicalizedParser;
import edu.stanford.nlp.process.PTBTokenizer;
import edu.stanford.nlp.util.Filter;
import edu.stanford.nlp.util.Pair;
import edu.stanford.nlp.util.StringUtils;

import java.io.*;
import java.util.*;

import static edu.stanford.nlp.trees.EnglishGrammaticalRelations.*;
import static edu.stanford.nlp.trees.GrammaticalRelation.*;

/**
 * A GrammaticalStructure for English.
 * <p/>
 * The parser should be run with the "-retainNPTmpSubcategories" option!
 * <b>Caveat emptor!</b> This is a work in progress.
 * Suggestions welcome.
 *
 * @author Bill MacCartney
 * @author Marie-Catherine de Marneffe
 */
public class EnglishGrammaticalStructure extends GrammaticalStructure {

  public static final String CONJ_MARKER = "conj_";

  private static final boolean DEBUG = false;

  //private static HeadFinder shf;

  /**
   * Construct a new <code>GrammaticalStructure</code> from an
   * existing parse tree.  The new <code>GrammaticalStructure</code>
   * has the same tree structure and label values as the given tree
   * (but no shared storage).  As part of construction, the parse tree
   * is analyzed using definitions from {@link GrammaticalRelation
   * <code>GrammaticalRelation</code>} to populate the new
   * <code>GrammaticalStructure</code> with as many labeled
   * grammatical relations as it can.
   */
  public EnglishGrammaticalStructure(Tree t) {
    this(t, new PennTreebankLanguagePack().punctuationWordRejectFilter());
  }

  /**
   * Construct a new <code>GrammaticalStructure</code> from an
   * existing parse tree.  The new <code>GrammaticalStructure</code>
   * has the same tree structure and label values as the given tree
   * (but no shared storage).  As part of construction, the parse tree
   * is analyzed using definitions from {@link GrammaticalRelation
   * <code>GrammaticalRelation</code>} to populate the new
   * <code>GrammaticalStructure</code> with as many labeled
   * grammatical relations as it can.
   */
  public EnglishGrammaticalStructure(Tree t, Filter<String> puncFilter) {
    super((new CoordinationTransformer()).transformTree(t), EnglishGrammaticalRelations.values(), getSemanticHeadFinder(t), puncFilter);
  }

  /**
   * Initialize shf according to the type of t we have (add flags?)
   * If t represents an existential sentence, we don't want a copular treatment of it.
   *
   * @param t
   */
  static protected SemanticHeadFinder getSemanticHeadFinder(Tree t) {
    SemanticHeadFinder shf;
    if (isExistential(t) || isWHQ(t)) {
      shf = new SemanticHeadFinder(false);
    } else {
      shf = new SemanticHeadFinder(true);
    }
    return shf;
  }


  static private boolean isExistential(Tree t) {
    boolean toReturn = false;
    List<Label> tags = t.preTerminalYield();
    for (Label tag : tags) {
      if (tag.value().equals("EX")) {
        toReturn = true;
      }
    }
    return toReturn;
  }

  /** Is the tree t a Wh-question?
   *  At present this is only true if the tree is headed by an SBARQ which has 
   *  both WH.* and SQ children.  (It was changed to looser definition in 
   *  Feb 2006. 
   */
  static private boolean isWHQ(Tree t) {
    if (t == null) return false;
    Tree firstChild = t.firstChild();
    if (firstChild == null) return false;
    boolean seenWH = false;
    boolean seenSQ = false;
    if ("SBARQ".equals(firstChild.value())) {
      Tree[] kids = firstChild.children();
      for (Tree kid : kids) {
        if (kid.value() != null) {
          if (kid.value().startsWith("WH")) {
            seenWH = true;
          } else if (kid.value().equals("SQ")) {
            seenSQ = true;
          }
        }
      }
    }
    // if (seenWH && seenSQ) System.err.println("It was a WH question!");
    return seenWH && seenSQ;
  }

  /**
   * Tries to return a node representing the <code>SUBJECT</code> (whether
   * nominal or clausal) of the given node <code>t</code>.
   * Probably, node <code>t</code> should represent a clause or
   * verb phrase.
   *
   * @param t a node in this <code>GrammaticalStructure</code>
   * @return a node which is the subject of node
   *         <code>t</code>, or else <code>null</code>
   */
  public TreeGraphNode getSubject(TreeGraphNode t) {
    TreeGraphNode subj = getNodeInRelation(t, NOMINAL_SUBJECT);
    if (subj == null) {
      return getNodeInRelation(t, CLAUSAL_SUBJECT);
    } else {
      return subj;
    }
  }

  protected void correctDependencies(Collection<TypedDependency> list) {
    correctSubjPassAndPoss(list);
  }

  private static void printListSorted(String title, 
                                      Collection<TypedDependency> list) {
    List<TypedDependency> lis = new ArrayList<TypedDependency>(list);
    Collections.sort(lis);
    if (title != null) System.err.println(title);
    System.err.println(lis);
  }


  /**
   * Destructively modifies this <code>TypedDependencyList</code> by
   * collapsing two types of transitive pairs of dependencies:
   * <dl>
   * <dt>prepositional dependencies</dt>
   * <dd>
   * <code>dep(cat, in)</code> and
   * <code>dep(in, hat)</code> are collapsed to
   * <code>in(cat, hat)</code>
   * </dd>
   * <dt>conjunct dependencies</dt>
   * <dd>
   * <code>cc(investors, and)</code> and
   * <code>conj(investors, regulators) are collapsed to
   * <code>and(investors,regulators)</code>
   * </dd>
   * </dl>
   * It will also erase possesive dependencies
   * <code>possessive(Montezuma, 's)</code> will be erased
   * For relative clauses, it will collapse referent such as:
   * <dd>
   * <code>ref(man, that)</code> and
   * <code>dobj(love, that)</code> are collapsed to
   * <code>dobj(love, man)</code>
   * </dd>
   */
  protected void collapseDependencies(Collection<TypedDependency> list, 
                                      boolean CCprocess) {
    if (DEBUG) printListSorted("collapseDependencies: CCproc: " + CCprocess, 
                               list);
    correctDependencies(list);
    if (DEBUG) printListSorted("After correctDependencies:", list);

    collapsePrepAndPoss(list);
    if (DEBUG) printListSorted("After PrepAndPoss:", list);

    collapseConj(list);
    if (DEBUG) printListSorted("After conj:", list);

    if (CCprocess) {
      treatCC(list);
    }
    if (DEBUG) printListSorted("After treatCC:", list);

    if (DEBUG) printListSorted("Before multiword preps:", list);
    collapseMultiwordPreps(list);

    if (DEBUG) printListSorted("Before collapse referent:", list);
    collapseReferent(list);

    Collections.sort((List) list);
    if (DEBUG) printListSorted("After all collapse:", list);
  }

  // some hard coding to deal with CONJP: for now we deal with
  // as well as, not to mention, rather than, but rather, instead of, but not, but also
  protected GrammaticalRelation conjValue(Object conj) {
    String newConj = conj.toString().toLowerCase();
    if (conj.toString().equals("not") || conj.toString().equals("instead") || conj.toString().equals("rather")) {
      newConj = "negcc";
    } else if (conj.toString().equals("to") || conj.toString().equals("also") || conj.toString().equals("well")) {
      newConj = "and";
    }
    //return CONJ_MARKER + newConj;
    return EnglishGrammaticalRelations.getConj(newConj);
  }

  private void treatCC(Collection<TypedDependency> list) {

    // Construct a map from tree nodes to the set of typed
    // dependencies in which the node appears as dependent.
    Map<TreeGraphNode, Set<TypedDependency>> map = new HashMap<TreeGraphNode, Set<TypedDependency>>();
    for (TypedDependency typedDep : list) {
      if (!map.containsKey(typedDep.dep())) {
        map.put(typedDep.dep(), new HashSet<TypedDependency>());
      }
      map.get(typedDep.dep()).add(typedDep);
    }

    Collection<TypedDependency> newTypedDeps = new ArrayList<TypedDependency>(list);

    // find typed deps of form conj(gov,dep)
    for (TypedDependency td : list) {
      if (EnglishGrammaticalRelations.getConjs().contains(td.reln())) {
      //if( td.reln().toString().startsWith("CC_")) {
        TreeGraphNode gov = td.gov();
        TreeGraphNode dep = td.dep();
        Set<TypedDependency> gov_relations = map.get(gov);
        if (gov_relations != null) {
          for (TypedDependency td1 : gov_relations) {
            //System.out.println("gov rel " + td1);
            TreeGraphNode newGov = td1.gov();
            GrammaticalRelation newRel = td1.reln();
            newTypedDeps.add(new TypedDependency(newRel, newGov, dep));
          }
        }
      }
    }
    list.clear();
    list.addAll(newTypedDeps);
  }

  private void collapseConj(Collection<TypedDependency> list) {
    // find typed deps of form cc(gov, dep)
    for (TypedDependency td : list) {
      if (td.reln() == COORDINATION) {
        TreeGraphNode gov = td.gov();
        GrammaticalRelation conj = conjValue(td.dep().value());

        //System.out.println("found cc " + td);

        // find other deps of that gov having reln "conj"
        for (TypedDependency td1 : list) {
          if (td1.reln() == CONJUNCT && td1.gov() == gov) {
            // and change "conj" to the actual (lexical) conjunction
            td1.setReln(conj);
          } else if (td1.reln() == COORDINATION) {
            conj = conjValue(td1.dep().value());
          }
        }
      }
    }
    // now remove typed dependencies with reln "cc"
    Collection<TypedDependency> newTypedDeps = new ArrayList<TypedDependency>();
    for (TypedDependency td : list) {
      if (td.reln() != COORDINATION) {
        newTypedDeps.add(td);
      }
    }
    list.clear();
    list.addAll(newTypedDeps);
  }

  // this method will collapse a referent relation such as follows
  // e.g.: "The man that I love ... "
  // ref(man, that)
  // dobj(love, that)
  // -> dobj(love, man)
  private void collapseReferent(Collection<TypedDependency> list) {
    // find typded deps of form ref(gov, dep)
    // put them in a List of Pairs
    List<Pair<TreeGraphNode, TreeGraphNode>> refs = new ArrayList<Pair<TreeGraphNode, TreeGraphNode>>();
    for (TypedDependency td : list) {
      if (td.reln() == REFERENT) {
        Pair<TreeGraphNode, TreeGraphNode> ref = new Pair<TreeGraphNode, TreeGraphNode>(td.dep(), td.gov());
        refs.add(ref);
      }
    }

    for (Pair<TreeGraphNode, TreeGraphNode> ref : refs) {
      TreeGraphNode dep = ref.first(); // take the relative word
      TreeGraphNode ant = ref.second(); // take the antecedent
      for (TypedDependency td : list) {
        if (td.dep() == dep && td.reln() != RELATIVE) {
          td.setDep(ant);
        }
      }
    }

    // now remove typed dependencies with reln "ref"
    Collection<TypedDependency> newTypedDeps = new ArrayList<TypedDependency>();
    for (TypedDependency td : list) {
      if (td.reln() != REFERENT) {
        newTypedDeps.add(td);
      }
    }
    list.clear();
    list.addAll(newTypedDeps);
  }

  private void correctSubjPassAndPoss(Collection<TypedDependency> list) {
    //put in a list verbs having an auxpass
    List<TreeGraphNode> list_auxpass = new ArrayList<TreeGraphNode>();
    for (TypedDependency td : list) {
      if (td.reln() == AUX_PASSIVE_MODIFIER) {
        list_auxpass.add(td.gov());
      }
    }
    //correct nsubj
    for (TypedDependency td : list) {
      if (td.reln() == NOMINAL_SUBJECT && list_auxpass.contains(td.gov())) {
        td.setReln(NOMINAL_PASSIVE_SUBJECT);
      }
      // correct unretrieved poss: dep relation in which the dependent is a
      // PRP$ or WP$
      // cdm jan 2006: couldn't we just recognize this in basic rules??
      String tag = td.dep().parent().value();
      if (td.reln() == DEPENDENT && (tag.equals("PRP$")||tag.equals("WP$"))) {
        td.setReln(POSSESSION_MODIFIER);
      }
    }
  }

  private void collapsePrepAndPoss(Collection<TypedDependency> list) {

    // Man oh man, how gnarly is the logic of this method.

    Collection<TypedDependency> newTypedDeps = new ArrayList<TypedDependency>();

    // Construct a map from tree nodes to the set of typed
    // dependencies in which the node appears as governor.
    // cdm: could use CollectionValuedMap here!
    Map<TreeGraphNode, Set<TypedDependency>> map = new HashMap<TreeGraphNode, Set<TypedDependency>>();
    for (TypedDependency typedDep : list) {
      if (!map.containsKey(typedDep.gov())) {
        map.put(typedDep.gov(), new HashSet<TypedDependency>());
      }
      map.get(typedDep.gov()).add(typedDep);
    }
    //System.err.println("here's the map: " + map);

    // Do preposition conjunction interaction for 
    // governor p NP and p NP case ... a lot of special code cdm jan 2006

    for (TypedDependency td1 : list) {
      if ( td1.reln() != KILL) {
        TreeGraphNode td1Dep = td1.dep();
        String td1DepPOS = td1Dep.parent().value();
        Set<TypedDependency> possibles = map.get(td1Dep);
        if (possibles == null) continue;
        // look for the "second half"
        TypedDependency prepDep = null;
        TypedDependency ccDep = null;
        TypedDependency conjDep = null;
        TypedDependency prep2Dep = null;
        Set<TypedDependency> otherDtrs = new HashSet<TypedDependency>();
        for (TypedDependency td2 : possibles) {
          if (td2.reln() == COORDINATION) {
            ccDep = td2;
          } else if (td2.reln() == CONJUNCT) {
            conjDep = td2;
            TreeGraphNode td2Dep = td2.dep();
            String td2DepPOS = td2Dep.parent().value();
            if ((td2DepPOS.equals("IN") || td2DepPOS.equals("TO")) &&
                td2Dep.value().equals(td1Dep.value())) {
              // Same preposition
              Set<TypedDependency> possibles2 = map.get(td2Dep);
              if (possibles2 == null) continue;
              for (TypedDependency td3 : possibles2) {
                TreeGraphNode td3Dep = td3.dep();
                String td3DepPOS = td3Dep.parent().value();
                // CDM Mar 2006: I put in disjunction here when I added in
                // PREPOSITIONAL_OBJECT.  If it catches all cases, we should
                // be able to delete the DEPENDENT disjunct
                // maybe better to delete the DEPENDENT disjunct want it creates problem with multiple prep (mcdm)
                if (td3.reln() == PREPOSITIONAL_OBJECT && prep2Dep == null &&
                    ( ! (td3DepPOS.equals("RB") || td3DepPOS.equals("IN") ||
                         td3DepPOS.equals("TO")))) {
                  prep2Dep = td3;
                } else {
                  otherDtrs.add(td3);
                }
              }
            }
          } else {
            TreeGraphNode td2Dep = td2.dep();
            String td2DepPOS = td2Dep.parent().value();
            // CDM Mar 2006: I put in disjunction here when I added in
            // PREPOSITIONAL_OBJECT.  If it catches all cases, we should
            // be able to delete the DEPENDENT disjunct
            if ((td1.reln() == PREPOSITIONAL_MODIFIER || td1.reln() == RELATIVE) && (td2.reln() == DEPENDENT || td2.reln() == PREPOSITIONAL_OBJECT) && (td1DepPOS.equals("IN") || td1DepPOS.equals("TO")) && prepDep == null &&
                ( ! (td2DepPOS.equals("RB") || td2DepPOS.equals("IN") ||
                     td2DepPOS.equals("TO")))) {
              prepDep = td2;
            } else {
              otherDtrs.add(td2);
            }
          }
        } // for td2 : possibles

        if (false && ccDep != null) {
          System.err.println("!!!!!");
          System.err.println("Kids of td1 are: " + possibles);
          System.err.println("prepDep: " + prepDep);
          System.err.println("ccDep: " + ccDep);
          System.err.println("conjDep: " + conjDep);
          System.err.println("prep2Dep: " + prep2Dep);
        }

        if (prepDep != null && ccDep != null && conjDep != null &&
            prep2Dep != null) {
          // OK, we have a conjunction over identical PPs
          // handling the case of an "agent": the governor of a "by" preposition must have an "auxpass" dependency
          // if it is the case, the "agent" variable becomes true
          boolean agent = false;
          if (td1Dep.value().equals("by")) {
            // look if we have an auxpass
            Set<TypedDependency> aux_pass_poss = map.get(td1.gov());
            if (aux_pass_poss != null) {
              for (TypedDependency td_pass : aux_pass_poss) {
                if (td_pass.reln() == AUX_PASSIVE_MODIFIER) {
                  agent = true;
                }
              }
            }
          }

          TypedDependency tdNew;
          if (agent) {
            tdNew = new TypedDependency(AGENT, td1.gov(), prepDep.dep());
            agent = false;
          } else {
            // for prepositions, use the preposition
            GrammaticalRelation reln;
            if (td1.reln() == RELATIVE) {
              reln = RELATIVE;
            } else {
              reln = EnglishGrammaticalRelations.getPrep(td1Dep.value().toLowerCase());
            }
            tdNew = new TypedDependency(reln, td1.gov(), prepDep.dep());
          }
          newTypedDeps.add(tdNew);
          TypedDependency tdNew2 = new TypedDependency(conjValue(ccDep.dep().value()), 
                                                       prepDep.dep(), prep2Dep.dep());
          newTypedDeps.add(tdNew2);
          if (DEBUG) System.err.println("ConjPP adding: " + tdNew + " " + tdNew2);
          td1.setReln(KILL);        // remember these are "used up"
          prepDep.setReln(KILL);
          ccDep.setReln(KILL);
          conjDep.setReln(KILL);
          prep2Dep.setReln(KILL);

          // promote dtrs that would be orphaned
          for (TypedDependency otd : otherDtrs) {
            otd.setGov(td1.gov());
          }
        }
      }
    } // for TypedDependency td1 : list

    // below here is the single preposition/possessor basic case!

    for (TypedDependency td1 : list) {
      boolean agent = false;
      if (td1.reln() != KILL) {
        TreeGraphNode td1Dep = td1.dep();
        String td1DepPOS = td1Dep.parent().value();
        //System.err.println("td1 & td1DepPOS: " + td1 + "&" + td1DepPOS);
        // find all other typedDeps having our dep as gov
        Set<TypedDependency> possibles = map.get(td1Dep);
        if (possibles != null) {

          //System.out.println("Possibles: " + possibles);
          // look for the "second half"
          for (TypedDependency td2 : possibles) {
            if (td2.reln() != COORDINATION && td2.reln() != CONJUNCT) {

              TreeGraphNode td2Dep = td2.dep();
              String td2DepPOS = td2Dep.parent().value();
              // CDM Mar 2006: I put in disjunction here when I added in
              // PREPOSITIONAL_OBJECT.  If it catches all cases, we should
              // be able to delete the DEPENDENT disjunct
              // maybe better to delete the DEPENDENT disjunct want it creates problem with multiple prep (mcdm)
              if ((td1.reln() == DEPENDENT || td1.reln() == PREPOSITIONAL_MODIFIER || td1.reln() == RELATIVE) &&  (td2.reln() == PREPOSITIONAL_OBJECT) && (td1DepPOS.equals("IN") || td1DepPOS.equals("TO")) &&
                  ( ! (td2DepPOS.equals("RB") || td2DepPOS.equals("IN") ||
                       td2DepPOS.equals("TO")))) {
                // OK, we have a pair td1, td2 to collapse to td3

                // handling the case of an "agent": the governor of a "by" preposition must have an "auxpass" dependency
                // if it is the case, the "agent" variable becomes true
                if (td1Dep.value().equals("by")) {
                  // look if we have an auxpass
                  Set<TypedDependency> aux_pass_poss = map.get(td1.gov());
                  if (aux_pass_poss != null) {
                    for (TypedDependency td_pass : aux_pass_poss) {
                      if (td_pass.reln() == AUX_PASSIVE_MODIFIER) {
                        agent = true;
                      }
                    }
                  }
                }

                TypedDependency td3;
                if (agent) {
                  td3 = new TypedDependency(AGENT, td1.gov(), td2.dep());
                  agent = false;
                } else {
                  // for prepositions, use the preposition
                  GrammaticalRelation reln;
                  if (td1.reln() == RELATIVE) {
                    reln = RELATIVE;
                  } else {
                    reln = EnglishGrammaticalRelations.getPrep(td1Dep.value().toLowerCase());
                  }
                  td3 = new TypedDependency(reln, td1.gov(), td2.dep());
                }
                if (DEBUG) System.err.println("PP adding: " + td3 + " deleting: " + td1 + " " + td2);
                newTypedDeps.add(td3);
                td1.setReln(KILL);        // remember these are "used up"
                td2.setReln(KILL);        // remember these are "used up"
              }
            }
          }

          // Now we need to see if there any TDs that will be "orphaned"
          // by this collapse.  Example: if we have:
          //   dep(drew, on)
          //   dep(on, book)
          //   dep(on, right)
          // the first two will be collapsed to on(drew, book), but then
          // the third one will be orphaned, since its governor no
          // longer appears.  So, change its governor to 'drew'.
          if (td1.reln() == KILL) {
            for (TypedDependency td2 : possibles) {
              if (td2.reln() != KILL && td2.reln() != COORDINATION && td2.reln() != CONJUNCT) {
                //System.err.println("td1 & td2: " + td1 + " & " + td2);
                td2.setGov(td1.gov());
              }
            }
          }
        }
      }
    }

    // now copy remaining unkilled TDs from here to new
    // POSSESSIVE_MODIFIERS should normally have no kids, as possessor is
    // already linked to posessee, but only delete it if there are really 
    // no other kids
    for (TypedDependency td : list) {
      boolean keep = true;
      if (td.reln() == KILL) {
        keep = false;
      } else if (td.reln() == POSSESSIVE_MODIFIER) {
        keep = false;
        TreeGraphNode dep = td.dep();
        // search for it
        for (TypedDependency typedD : list) {
          if (typedD.gov().equals(dep)) {
            keep = true;
            break;
          }
        }
      }
      if (keep) {
        newTypedDeps.add(td);
      }
    } // for TypedDependency td

    list.clear();                            // forget all (esp. killed) TDs
    list.addAll(newTypedDeps);
  } // end collapsePrepAndPoss()


  // used by collapseMultiwordPreps()  KEPT IN ALPHABETICAL ORDER
  private static String[][] MULTIWORD_PREPS = {{"according", "to"}, {"across", "from"}, // may not work (across parsed as particle)
                                               {"ahead", "of"}, // may not work (ahead parsed as adverb)
                                               {"along", "with"}, {"due", "to"}, {"alongside", "of"}, {"apart", "from"}, {"as", "of"}, {"as", "to"}, {"away", "from"}, {"based", "on"}, {"because", "of"}, {"close", "by"}, {"close", "to"}, {"due", "to"}, {"compared", "to"}, {"compared", "with"}, {"depending", "on"}, {"followed", "by"}, {"inside", "of"}, {"instead", "of"}, {"next", "to"}, {"near", "to"}, {"out", "of"}, {"outside", "of"}, {"prior", "to"}, {"together", "with"}};

  private void collapseMultiwordPreps(Collection<TypedDependency> list) {
    Collection<TypedDependency> newTypedDeps = new ArrayList<TypedDependency>();

    for (String[] mwp : MULTIWORD_PREPS) {
      // find typed deps of form dep(gov, mwp[0])
      for (TypedDependency td1 : list) {
        if (td1.dep().value().equalsIgnoreCase(mwp[0])) {
          // find other deps of that dep having reln prep_mwp[1]
          for (TypedDependency td2 : list) {
            if ((td2.gov() == td1.dep() || td2.gov() == td1.gov()) && td2.reln().getSpecific() != null && td2.reln().getSpecific().equals(mwp[1])) {
              td1.setReln(KILL);
              td2.setReln(KILL);
              GrammaticalRelation gr = EnglishGrammaticalRelations.getPrep(mwp[0] + "_" + mwp[1]);
              newTypedDeps.add(new TypedDependency(gr, td1.gov(), td2.dep()));
            }
          }
        }
      }
    }

    // now remove typed dependencies with reln "kill"
    for (TypedDependency td : list) {
      if (td.reln() != KILL) {
        newTypedDeps.add(td);
      }
    }
    list.clear();
    list.addAll(newTypedDeps);
  }

  /**
   * Just for testing.
   * Usage: <br> <code>
   * java edu.stanford.nlp.trees.EnglishGrammaticalStructure -[treeFile FILE | -sentFile FILE] <br>
   *      [-testGraph GRAPH]</code>
   */
  public static void main(String[] args) {

    // System.out.print("GrammaticalRelations under DEPENDENT:");
    // System.out.println(DEPENDENT.toPrettyString());

    Treebank tb = new MemoryTreebank();
    Properties props = StringUtils.argsToProperties(args);
    String treeFileName = props.getProperty("treeFile");
    String sentFileName = props.getProperty("sentFile");

    if (sentFileName == null && treeFileName == null) {
      try {
        System.err.println("Usage: java EnglishGrammaticalStructure [-sentFile file|-treeFile file] [-testGraph]");
        TreeReader tr = new PennTreeReader(new StringReader("((S (NP (NNP Sam)) (VP (VBD died) (NP-TMP (NN today)))))"), new LabeledScoredTreeFactory());
        tb.add(tr.readTree());
      } catch (Exception e) {
        System.err.println("Horrible error: " + e);
        e.printStackTrace();
      }
    } else if (treeFileName != null) {
        tb.loadPath(treeFileName);
    } else {
      String[] opts = new String[]{"-retainNPTmpSubcategories"};
      LexicalizedParser lp = new LexicalizedParser("/u/nlp/data/lexparser/englishPCFG.ser.gz");
      lp.setOptionFlags(opts);
      BufferedReader reader = null;
      try {
        reader = new BufferedReader(new FileReader(sentFileName));
      } catch (FileNotFoundException e) {
        System.err.println("Cannot find " + sentFileName);
        System.exit(1);
      }
      try {
        System.out.println("Processing sentence file " + sentFileName);
        String line;
        while ((line = reader.readLine()) != null) {
          System.out.println("Processing sentence: "  + line);
          PTBTokenizer ptb = new PTBTokenizer(new StringReader(line));
          List words = ptb.tokenize();
          lp.parse(words);
          Tree parseTree = lp.getBestParse();
          tb.add(parseTree);
        }
        reader.close();
      } catch (Exception e) {
        e.printStackTrace();
        System.err.println("IOexception reading key file " + sentFileName);
        System.exit(1);
      }
    }

    System.out.println("Phrase structure tree, then dependencies, then collapsed dependencies");
    for (Tree t : tb) {

      System.out.println("==================================================");
      GrammaticalStructure gs = new EnglishGrammaticalStructure(t);

      t.pennPrint();

      //System.out.println("----------------------------");
      //TreeGraph tg = new TreeGraph(t);
      //System.out.println(tg);

      System.out.println("----------------------------");
      System.out.println(gs);

      System.out.println("----------------------------");
      System.out.println(StringUtils.join(gs.typedDependencies(true), "\n"));

      System.out.println("----------------------------");
      System.out.println(StringUtils.join(gs.typedDependenciesCollapsed(true), "\n"));

      System.out.println("----------------------------");
      System.out.println(StringUtils.join(gs.typedDependenciesCCprocessed(false), "\n"));

      //gs.printTypedDependencies("xml");

      // cdm May 2006: this used to make and test the SemanticGraph, but
      // I've moved that code to SemanticGraph to delink things.
    }
  } // end main

}

