package edu.stanford.nlp.trees;

import edu.stanford.nlp.util.Timing;
import edu.stanford.nlp.parser.lexparser.TreebankLangParserParams;
import edu.stanford.nlp.parser.lexparser.EnglishTreebankParserParams;
import edu.stanford.nlp.ling.Label;

import java.io.File;
import java.io.FileFilter;
import java.io.Reader;
import java.util.Iterator;

/**
 * @author Pi-Chuan Chang
 */
public class TransformingTreebank extends Treebank {
  TreeTransformer transformer;
  Treebank tf;
  Iterator iter = null;

  /**
   * Create a new TransformingTreebank.
   * The trees are made with a <code>LabeledScoredTreeReaderFactory</code>.
   * <p/>
   * <i>Compatibility note: Until Sep 2004, this used to create a Treebank
   * with a SimpleTreeReaderFactory, but this was changed as the old
   * default wasn't very useful, especially to naive users.</i>
   */
  public TransformingTreebank() {
    this(new LabeledScoredTreeReaderFactory());
  }

  /**
   * Create a new tree bank, set the encoding for file access.
   *
   * @param encoding The charset encoding to use for treebank file decoding
   */
  public TransformingTreebank(String encoding) {
    this(new LabeledScoredTreeReaderFactory(), encoding);
  }

  /**
   * Create a new TransformingTreebank.
   *
   * @param trf the factory class to be called to create a new
   *            <code>TreeReader</code>
   */
  public TransformingTreebank(TreeReaderFactory trf) {
    super(trf);
  }

  /**
   * Create a new TransformingTreebank.
   *
   * @param trf      the factory class to be called to create a new
   *                 <code>TreeReader</code>
   * @param encoding The charset encoding to use for treebank file decoding
   */
  public TransformingTreebank(TreeReaderFactory trf, String encoding) {
    super(trf, encoding);
  }




  public TransformingTreebank(Treebank tf, TreeTransformer transformer) {
    this.tf = tf;
    this.transformer = transformer;
  }

  /**
   * Empty a <code>Treebank</code>.
   */
  public void clear() {
    tf.clear();
    transformer = null;
  };

  /**
   * Load trees from given path specification.
   *
   * @param path file or directory to load from
   * @param filt a FilenameFilter of files to load
   */
  public void loadPath(File path, FileFilter filt) {
    throw new UnsupportedOperationException();
  };

  /**
   * Applies the TreeVisitor to to all trees in the Treebank.
   *
   * @param tp A class that can process trees.
   */
  public void apply(TreeVisitor tp) {
    for (Tree t : this.tf) {
      Tree tmpT = t.deeperCopy();
      if (transformer != null) {
        tmpT = transformer.transformTree(tmpT);
        //System.err.println("Transform!!!! in apply");
      }
      tp.visitTree(tmpT);
    }
  }

  /**
   */
  public Iterator<Tree> iterator() {
    if (iter == null) {
      iter = new TransformingTreebankIterator(tf.iterator());
    }
    return iter;
  }

  /**
   * Loads treebank grammar from first argument and prints it.
   * Just a demonstration of functionality. <br>
   * <code>usage: java MemoryTreebank treebankFilesPath</code>
   *
   * @param args array of command-line arguments
   */
  public static void main(String[] args) {
    Timing.startTime();
    Treebank treebank = new DiskTreebank(new TreeReaderFactory() {
      public TreeReader newTreeReader(Reader in) {
        return new PennTreeReader(in, new LabeledScoredTreeFactory());
      }
    });
    Treebank treebank2 = new MemoryTreebank(new TreeReaderFactory() {
      public TreeReader newTreeReader(Reader in) {
        return new PennTreeReader(in, new LabeledScoredTreeFactory());
      }
    });
    treebank.loadPath(args[0]);
    treebank2.loadPath(args[0]);
    CompositeTreebank c = new CompositeTreebank(treebank, treebank2);
    Timing.endTime();
    TreeTransformer myTransformer = new MyTreeTransformer();
    TreeTransformer myTransformer2 = new MyTreeTransformer2();
    TreeTransformer myTransformer3 = new MyTreeTransformer3();
    //Treebank tf = new TransformingTreebank(new TransformingTreebank(new TransformingTreebank(c, myTransformer), myTransformer2), myTransformer3);
    Treebank tf = c.transform(myTransformer).transform(myTransformer2).transform(myTransformer3);

    System.out.println(c);
    System.out.println("-------------------------");
    System.out.println(tf);
    System.out.println("-------------------------");
  }


  private class TransformingTreebankIterator implements Iterator<Tree> {
    Iterator<Tree> iter = null;
    TransformingTreebankIterator (Iterator<Tree> iter) {
      this.iter = iter;
    }
    public boolean hasNext() {
      return iter.hasNext();
    }
    public Tree next() {
      if (iter.hasNext()) {
        Tree ret = iter.next();
        if (transformer != null) {
          ret = transformer.transformTree(ret);
          //System.err.println("Transform!!!! in next");
        }
        return ret;
      } else {
        return null;
      }
    }
    public void remove() {
      throw new UnsupportedOperationException();
    }
  }

}


class MyTreeTransformer implements TreeTransformer {

  public Tree transformTree(Tree tree) {
    Tree treeCopy = tree.deepCopy();
    for (Iterator it = treeCopy.iterator(); it.hasNext();) {
      Tree subtree = (Tree) it.next();
      if (subtree.depth() < 2) {
        continue;
      }
      String categoryLabel = subtree.label().toString();
      Label label = subtree.label();
      label.setFromString(categoryLabel+"-t1");
    }
    return treeCopy;
  }
}


class MyTreeTransformer2 implements TreeTransformer {

  public Tree transformTree(Tree tree) {
    Tree treeCopy = tree.deepCopy();
    for (Iterator it = treeCopy.iterator(); it.hasNext();) {
      Tree subtree = (Tree) it.next();
      if (subtree.depth() < 2) {
        continue;
      }
      String categoryLabel = subtree.label().toString();
      Label label = subtree.label();
      label.setFromString(categoryLabel+"-t2");
    }
    return treeCopy;
  }
}


class MyTreeTransformer3 implements TreeTransformer {

  public Tree transformTree(Tree tree) {
    Tree treeCopy = tree.deepCopy();
    for (Iterator it = treeCopy.iterator(); it.hasNext();) {
      Tree subtree = (Tree) it.next();
      if (subtree.depth() < 2) {
        continue;
      }
      String categoryLabel = subtree.label().toString();
      Label label = subtree.label();
      label.setFromString(categoryLabel+"-t3");
    }
    return treeCopy;
  }
}
