package edu.stanford.nlp.trees;

import edu.stanford.nlp.ling.*;
import edu.stanford.nlp.parser.lexparser.TreebankLangParserParams;
import edu.stanford.nlp.process.Function;
import edu.stanford.nlp.trees.international.pennchinese.ChineseEnglishWordMap;
import edu.stanford.nlp.util.Filter;
import edu.stanford.nlp.util.Filters;
import edu.stanford.nlp.util.StringUtils;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.*;


/**
 * A class for customizing the print method(s) for a
 * <code>edu.stanford.nlp.trees.Tree</code> as the output of the
 * parser.  This class supports printing in multiple ways and altering
 * behavior via properties. 
 *
 * @author Roger Levy
 * @author Christopher Manning
 * @author Galen Andrew
 */
public class TreePrint {

  public static final String rootLabelOnlyFormat = "rootSymbolOnly";

  /** The legal output tree formats. */
  public static final String[] outputTreeFormats = {
    "penn",
    "oneline",
    rootLabelOnlyFormat,
    "words",
    "wordsAndTags",
    "dependencies",
    "typedDependencies",
    "typedDependenciesCollapsed",
    "latexTree",
    "collocations" 
  };

  public static final String headMark = "=H";

  private Properties formats;
  private Properties options;

  private boolean markHeadNodes = false;
  private boolean lexicalize = false;
  private boolean stem = false;
  private boolean transChinese = false;

  private HeadFinder hf;
  private TreebankLanguagePack tlp;
  private WordStemmer stemmer;
  private Filter<Dependency> dependencyFilter;
  private GrammaticalStructureFactory gsf;

  /** This PrintWriter is used iff the user doesn't pass one in to a 
   *  call to printTree().
   */
  private PrintWriter pw = new PrintWriter(System.out, true);

  /** The anglocentric constructor */
  public TreePrint(String formats) {
    this(formats, "", new PennTreebankLanguagePack());
  }

  public TreePrint(String formats, TreebankLanguagePack tlp) {
    this(formats, "", tlp);
  }

  public TreePrint(String formats, String options, TreebankLanguagePack tlp) {
    this(formats, options, tlp, new CollinsHeadFinder());
  }

  /**
   * Make a TreePrint instance.
   *
   * @param formatString A comma separated list of ways to print each Tree.
   *                Known formats are: oneline, penn, latexTree, words, 
   *                wordsAndTags, rootSymbolOnly, dependencies,
   *                typedDependencies, typedDependenciesCollapsed, 
   *                collocations.  All of them print a blank line after
   *                the output except for oneline.  oneline is also not 
   *                meaningul in XML output (it is ignored: use penn instead).
   * @param optionsString Options that additionally specify how trees are to
   *                be printed (for instance, whether stemming should be done).
   *                Known options are: <code>stem, lexicalize, markHeadNodes,
   *                xml, removeTopBracket, transChinese,
   *                includePunctuationDependencies
   *                </code>.
   * @param tlp     The TreebankLanguagePack used to do things like delete
   *                or ignore punctuation in output
   * @param hf      The HeadFinder used in printing output
   */
  public TreePrint(String formatString, String optionsString, TreebankLanguagePack tlp, HeadFinder hf) {
    formats = StringUtils.stringToProperties(formatString);
    options = StringUtils.stringToProperties(optionsString);
    List<String> okOutputs = Arrays.asList(outputTreeFormats);
    for (Object formObj : formats.keySet()) {
      String format = (String) formObj;
      if ( ! okOutputs.contains(format)) {
        throw new RuntimeException("Error: output tree format " + format + " not supported");
      }
    }

    setHeadFinder(hf);
    this.tlp = tlp;
    boolean includePunctuationDependencies;
    includePunctuationDependencies = propertyToBoolean(this.options,
                                                       "includePunctuationDependencies");
    Filter<String> puncWordFilter;
    if (includePunctuationDependencies) {
      dependencyFilter = Filters.acceptFilter();
      puncWordFilter = Filters.acceptFilter();
    } else {
      dependencyFilter = new DependencyRejectFilter(tlp.punctuationTagRejectFilter());
      puncWordFilter = tlp.punctuationWordRejectFilter();
    }
    stem = propertyToBoolean(this.options, "stem");
    if (stem) {
      stemmer = new WordStemmer();
    }
    if (formats.containsKey("typedDependenciesCollapsed") || formats.containsKey("typedDependencies")) {
      gsf = tlp.grammaticalStructureFactory(puncWordFilter);
    }

    lexicalize = propertyToBoolean(this.options, "lexicalize");
    markHeadNodes = propertyToBoolean(this.options, "markHeadNodes");
    transChinese = propertyToBoolean(this.options, "transChinese");
  }


  private static boolean propertyToBoolean(Properties prop, String key) {
    String val = prop.getProperty(key);
    if (val == null) {
      return false;
    } else {
      val = val.trim();
      if (val.equals("")) {
        return true;
      } else if (val.equalsIgnoreCase("true")) {
        return true;
      } else if (val.equalsIgnoreCase("false")) {
        return false;
      } else {
        throw new RuntimeException("Property " + key + " should have a boolean value but it's " + val);
      }
    }
  }


  /**
   * Prints the tree to the default PrintWriter.
   */
  public void printTree(Tree t) {
    printTree(t, pw);
  }

  /**
   * Prints the tree, with an empty ID.
   */
  public void printTree(final Tree t, PrintWriter pw) {
    printTree(t, "", pw);
  }

  /**
   * Prints the tree according to the options specified for this instance.
   * If the tree <code>t</code> is <code>null</code>, then the code prints
   * a line indicating a skipped tree.  Under the XML option this is
   * an <code>s</code> element with the <code>skipped</code> attribute having
   * value <code>true</code>, and, otherwise, it is the token
   * <code>SENTENCE_SKIPPED_OR_UNPARSABLE</code>.
   *
   * @param id A name for this sentence.
   */
  public void printTree(final Tree t, final String id, final PrintWriter pw) {
    if (t == null) {
      // Parsing didn't succeed.
      if (propertyToBoolean(options, "xml")) {
        pw.print("<s");
        if (id != null && !"".equals(id)) {
          pw.print(" id=\"" + id + "\"");
        }
        pw.println(" skipped=\"true\"/>");
        pw.println();
      } else {
        pw.println("SENTENCE_SKIPPED_OR_UNPARSABLE");
      }
      return;
    }

    Tree outputTree = t;
    if (propertyToBoolean(options, "xml")) {
      Sentence sentUnstemmed = outputTree.yield();
      pw.print("<s");
      if (id != null && !"".equals(id)) {
        pw.print(" id=\"" + id + "\"");
      }
      pw.println(">");
      if (formats.containsKey("words")) {
        pw.println("  <words>");
        int i = 1;
        for (HasWord w : sentUnstemmed) {
          pw.println("    <word ind=\"" + i + "\">" + w + "</word>");
          i++;
        }
        pw.println("  </words>");
      }
    } else {
      if (formats.containsKey("words")) {
        pw.println(outputTree.yield().toString(false));
        pw.println();
      }
    }

    if (propertyToBoolean(options, "removeTopBracket")) {
      String s = outputTree.label().value();
      if (tlp.isStartSymbol(s)) {
        if (outputTree.isUnaryRewrite()) {
          outputTree = outputTree.firstChild();
        } else {
          // It's not quite clear what to do if the tree isn't unary at the top
          // but we then don't strip the ROOT symbol, since that seems closer
          // than losing part of the tree altogether....
          System.err.println("TreePrint: can't remove top bracket: not unary");
        }
      }
      // Note that TreePrint is also called on dependency trees that have
      // a word as the root node, and so we don't error if there isn't
      // the root symbol at the top; rather we silently assume that this
      // is a dependency tree!!
    }
    if (stem) {
      stemmer.visitTree(outputTree);
    }
    if (lexicalize) {
      Function<Tree, Tree> a = 
        TreeFunctions.getLabeledTreeToCategoryWordTagTreeFunction();
      outputTree = a.apply(outputTree);
      outputTree.percolateHeads(hf);
    }

    if (formats.containsKey("collocations")) {
      /* Reflection to check that a Wordnet connection exists.  Otherwise
       * we print an error and exit.
       */
      try {
        Class cl = Class.forName("edu.stanford.nlp.trees.WordNetInstance");
        WordNetConnection wn = (WordNetConnection) cl.newInstance();
        CollocationFinder cf = new CollocationFinder(outputTree, wn, hf);
        outputTree = cf.getMangledTree();
      } catch (Exception e) {
        e.printStackTrace();
        System.err.println("Couldn't open WordNet Connection.  Aborting collocation detection.");
      }
    }

    if (!lexicalize) { // delexicalize the output tree
      Function<Tree, Tree> a = 
        TreeFunctions.getLabeledTreeToStringLabeledTreeFunction();
      outputTree = a.apply(outputTree);
    }

    Tree outputPSTree = outputTree;  // variant with head-marking, translations

    if (markHeadNodes) {
      outputPSTree = markHeadNodes(outputPSTree);
    }

    if (transChinese) {
      TreeTransformer tt = new TreeTransformer() {
        public Tree transformTree(Tree t) {
          t = t.deepCopy();
          for (Tree subtree : t) {
            if (subtree.isLeaf()) {
              Label oldLabel = subtree.label();
              String translation = ChineseEnglishWordMap.getStaticMap().getFirstTranslation(oldLabel.value());
              if (translation == null) translation = "[UNK]";
              Label newLabel = new StringLabel(oldLabel.value()+":"+translation);
              subtree.setLabel(newLabel);
            }
          }
          return t;
        }
      };
      outputPSTree = tt.transformTree(outputPSTree);
    }

    if (propertyToBoolean(options, "xml")) {
      if (formats.containsKey("wordsAndTags")) {
        Sentence sent = outputTree.taggedYield();
        pw.println("  <words pos=\"true\">");
        int i = 1;
        for (Iterator it = sent.iterator(); it.hasNext();) {
          TaggedWord tw = (TaggedWord) it.next();
          pw.println("    <word ind=\"" + i + "\" pos=\"" + tw.tag() + "\">" + tw.word() + "</word>");
          i++;
        }
        pw.println("  </words>");
      }
      if (formats.containsKey("penn")) {
        pw.println("  <tree style=\"penn\">");
        outputPSTree.pennPrint(pw);
        pw.println("  </tree>");
      }
      if (formats.containsKey("latexTree")) {
        pw.println("    <tree style=\"latexTrees\">");
        pw.println(".[");
        outputTree.indentedListPrint(pw,false);
        pw.println(".]");
        pw.println("  </tree>");
      }
      if (formats.containsKey("dependencies")) {
        Tree indexedTree = outputTree.deeperCopy(outputTree.treeFactory(),
                                                 MapLabel.factory());
        indexedTree.indexLeaves();
        Set<Dependency> depsSet = indexedTree.mapDependencies(dependencyFilter, hf);
        List<Dependency> sortedDeps = new ArrayList<Dependency>(depsSet);
        Collections.sort(sortedDeps, ComparatorHolder.dc);
        pw.println("<dependencies style=\"untyped\">");
        for (Dependency d : sortedDeps) {
          pw.println(d.toString("xml"));
        }
        pw.println("</dependencies>");
      }
      if (formats.containsKey("typedDependencies")) {
        GrammaticalStructure gs = gsf.newGrammaticalStructure(outputTree);
        print(gs.typedDependencies(), "xml", pw);
      }
      if (formats.containsKey("typedDependenciesCollapsed")) {
        GrammaticalStructure gs = gsf.newGrammaticalStructure(outputTree);
        print(gs.typedDependenciesCollapsed(), "xml", pw);
      }
      pw.println("</s>");
      pw.println();
    } else {
      if (formats.containsKey("wordsAndTags")) {
        pw.println(outputTree.taggedYield().toString(false));
        pw.println();
      }
      if (formats.containsKey("oneline")) {
        pw.println(outputTree.toString());
      }
      if (formats.containsKey("penn")) {
        outputPSTree.pennPrint(pw);
        pw.println();
      }
      if (formats.containsKey(rootLabelOnlyFormat)) {
        pw.println(outputTree.label());
      }
      if (formats.containsKey("latexTree")) {
        pw.println(".[");
        outputTree.indentedListPrint(pw,false);
        pw.println(".]");
      }
      if (formats.containsKey("dependencies")) {
        Tree indexedTree = outputTree.deeperCopy(outputTree.treeFactory(),
                                                 MapLabel.factory());
        indexedTree.indexLeaves();
        Set<Dependency> depsSet = indexedTree.mapDependencies(dependencyFilter, hf);
        List<Dependency> sortedDeps = new ArrayList<Dependency>(depsSet);
        Collections.sort(sortedDeps, ComparatorHolder.dc);
        for (Dependency d : sortedDeps) {
          pw.println(d.toString("predicate"));
        }
        pw.println();
      }
      if (formats.containsKey("typedDependencies")) {
        // pw.println(new GrammaticalStructure(outputTree).typedDependencies());
        GrammaticalStructure gs = gsf.newGrammaticalStructure(outputTree);
        // pw.print(gs);
        print(gs.typedDependencies(), pw);
        pw.println();
      }
      if (formats.containsKey("typedDependenciesCollapsed")) {
        GrammaticalStructure gs = gsf.newGrammaticalStructure(outputTree);
        // pw.print(gs);
        print(gs.typedDependenciesCollapsed(), pw);
      }

    }
  }


  public void printHeader(PrintWriter pw, String charset) {
    if (propertyToBoolean(options, "xml")) {
      pw.println("<?xml version=\"1.0\" encoding=\"" + charset + "\"?>");
      pw.println("<corpus>");
    }
  }


  public void printFooter(PrintWriter pw) {
    if (propertyToBoolean(options, "xml")) {
      pw.println("</corpus>");
    }
  }


  /**
   * Sets whether or not to stem the Trees using Morphology.
   */
  public void setStem(boolean stem) {
    this.stem = stem;
    if (stem) {
      stemmer = new WordStemmer();
    } else {
      stemmer = null;
    }
  }

  /**
   * Set the headfinder to be used for lexicalizing trees.  Relevant
   * if the output is to be lexicalized.  By default,
   * <code>CollinsHeadFinder</code> is used.
   */
  public void setHeadFinder(HeadFinder hf) {
    this.hf = hf;
  }

  public HeadFinder getHeadFinder() {
    return hf;
  }


  /**
   * Sets the default print writer for printing trees with the instance.
   * Unless this is set, the default is <code>System.out</code>.
   */
  public void setPrintWriter(PrintWriter pw) {
    this.pw = pw;
  }

  public PrintWriter getPrintWriter() {
    return pw;
  }

  public Tree markHeadNodes(Tree t) {
    return markHeadNodes(t, null);
  }

  private Tree markHeadNodes(Tree t, Tree head) {
    if (t.isLeaf()) {
      return t; // don't worry about head-marking leaves
    }
    Label newLabel;
    if (t == head) {
      newLabel = headMark(t.label());
    } else {
      newLabel = t.label();
    }
    Tree newHead = hf.determineHead(t);
    return t.treeFactory().newTreeNode(newLabel, Arrays.asList(headMarkChildren(t, newHead)));
  }

  private Label headMark(Label l) {
    Label l1 = l.labelFactory().newLabel(l);
    l1.setValue(l1.value() + headMark);
    return l1;
  }

  private Tree[] headMarkChildren(Tree t, Tree head) {
    Tree[] kids = t.children();
    Tree[] newKids = new Tree[kids.length];
    for (int i = 0, n = kids.length; i < n; i++) {
      newKids[i] = markHeadNodes(kids[i], head);
    }
    return newKids;
  }

  /**
   * utility code...
   */
  public static void main(String[] args) {
    String format = "penn";
    String options = "";
    String tlppName = "edu.stanford.nlp.parser.lexparser.EnglishTreebankParserParams";
    String hfName = null;
    Map<String,Integer> flagMap = new HashMap<String,Integer>();
    flagMap.put("-format", 1);
    flagMap.put("-options", 1);
    flagMap.put("-tLPP", 1);
    flagMap.put("-hf", 1);
    Map<String,String[]> argsMap = StringUtils.argsToMap(args,flagMap);
    args = argsMap.get(null);
    if(argsMap.keySet().contains("-format")) {
      format = argsMap.get("-format")[0];
    }
    if(argsMap.keySet().contains("-options")) {
      options = argsMap.get("-options")[0];
    }
    if (argsMap.keySet().contains("-tLPP")) {
      tlppName = argsMap.get("-tLPP")[0];
    }
    if (argsMap.keySet().contains("-hf")) {
      hfName = argsMap.get("-hf")[0];
    }
    TreebankLangParserParams tlpp;
    try {
      tlpp = (TreebankLangParserParams) Class.forName(tlppName).newInstance();
    } catch (Exception e) {
      e.printStackTrace();
      return;
    }
    HeadFinder hf;
    try {
      hf = (HeadFinder) Class.forName(hfName).newInstance();
    } catch (Exception e) {
      e.printStackTrace();
      return;
    }
    TreePrint print = new TreePrint(format, options, tlpp.treebankLanguagePack(), (hf == null) ? tlpp.headFinder(): hf);
    Iterator<Tree> i; // initialized below
    if(args.length > 0) {
      Treebank trees; // initialized below
      if(argsMap.keySet().contains("-arabic"))
        trees = tlpp.diskTreebank();
      else {
        trees = new DiskTreebank(new TreeReaderFactory() {
            public TreeReader newTreeReader(Reader in) {
              return new PennTreeReader(in, new LabeledScoredTreeFactory(new StringLabelFactory()), new TreeNormalizer());
            }
          });
      }
      trees.loadPath(args[0]);
      i = trees.iterator();
    }
    else {
      i = tlpp.treeTokenizerFactory().getIterator(new BufferedReader(new InputStreamReader(System.in)));
    }
    while(i.hasNext()) {
      print.printTree(i.next());
    }
  }

  /**
   * NO OUTSIDE USE
   * Returns a String representation of the result of this set of
   * typed dependencies in a user-specified format.
   * Currently, three formats are supported:
   * <dl>
   * <dt>"plain"</dt>
   * <dd>(Default.)  Formats the dependencies as logical relations,
   * as exemplified by the following:
   * <pre>
   *  nsubj(died-1, Sam-0)
   *  tmod(died-1, today-2)
   *  </pre>
   * </dd>
   * <dt>"readable"</dt>
   * <dd>Formats the dependencies as a table with columns
   * <code>dependent</code>, <code>relation</code>, and
   * <code>governor</code>, as exemplified by the following:
   * <pre>
   *  Sam-0               nsubj               died-1
   *  today-2             tmod                died-1
   *  </pre>
   * </dd>
   * <dt>"xml"</dt>
   * <dd>Formats the dependencies as XML, as exemplified by the following:
   * <pre>
   *  &lt;dependencies&gt;
   *    &lt;dep type="nsubj"&gt;
   *      &lt;governor idx="1"&gt;died&lt;/governor&gt;
   *      &lt;dependent idx="0"&gt;Sam&lt;/dependent&gt;
   *    &lt;/dep&gt;
   *    &lt;dep type="tmod"&gt;
   *      &lt;governor idx="1"&gt;died&lt;/governor&gt;
   *      &lt;dependent idx="2"&gt;today&lt;/dependent&gt;
   *    &lt;/dep&gt;
   *  &lt;/dependencies&gt;
   *  </pre>
   * </dd>
   * </dl>
   *
   * @param format a <code>String</code> specifying the desired format
   * @return a <code>String</code> representation of the typed
   *         dependencies in this <code>GrammaticalStructure</code>
   */
  public static String toString(Collection<TypedDependency> dependencies, String format) {
    if (format != null && format.equals("xml")) {
      return toXMLString(dependencies);
    } else if (format != null && format.equals("readable")) {
      return toReadableString(dependencies);
    } else {
      return toString(dependencies);
    }
  }

  /**
   * NO OUTSIDE USE
   * Returns a String representation of this set of typed dependencies
   * as exemplified by the following:
   * <pre>
   *  tmod(died-6, today-9)
   *  nsubj(died-6, Sam-3)
   *  </pre>
   *
   * @return a <code>String</code> representation of this set of
   *         typed dependencies
   */
  public static String toString(Collection<TypedDependency> dependencies) {
    StringBuilder buf = new StringBuilder();
    for (TypedDependency td : dependencies) {
      buf.append(td.toString()).append("\n");
    }
    return buf.toString();
  }

  // NO OUTSIDE USE
  private static String toReadableString(Collection<TypedDependency> dependencies) {
    StringBuilder buf = new StringBuilder();
    buf.append(String.format("%-20s%-20s%-20s%n", "dep", "reln", "gov"));
    buf.append(String.format("%-20s%-20s%-20s%n", "---", "----", "---"));
    for (TypedDependency td : dependencies) {
      buf.append(String.format("%-20s%-20s%-20s%n", td.dep(), td.reln(), td.gov()));
    }
    return buf.toString();
  }

  // NO OUTSIDE USE
  private static String toXMLString(Collection<TypedDependency> dependencies) {
    StringBuilder buf = new StringBuilder("<dependencies style=\"typed\">\n");
    for (TypedDependency td : dependencies) {
      String reln = td.reln().toString();
      String gov = td.gov().value();
      int govIdx = td.gov().index();
      String dep = td.dep().value();
      int depIdx = td.dep().index();
      buf.append("  <dep type=\"" + reln + "\">\n");
      buf.append("    <governor idx=\"" + govIdx + "\">" + gov + "</governor>\n");
      buf.append("    <dependent idx=\"" + depIdx + "\">" + dep + "</dependent>\n");
      buf.append("  </dep>\n");
    }
    buf.append("</dependencies>");
    return buf.toString();
  }

  /**
   * USED BY TREEPRINT AND PREPROCESS
   * Prints this set of typed dependencies to the specified
   * <code>PrintWriter</code>.
   */
  public static void print(Collection<TypedDependency> dependencies, PrintWriter pw) {
    pw.println(toString(dependencies));
  }

  /**
   * USED BY TREEPRINT
   * Prints this set of typed dependencies to the specified
   * <code>PrintWriter</code> in the specified format.
   */
  public static void print(Collection<TypedDependency> dependencies, String format, PrintWriter pw) {
    pw.println(toString(dependencies, format));
  }

  private static class DependencyRejectFilter implements Filter<Dependency> {

    private Filter tagRejectFilter;

    DependencyRejectFilter(Filter trf) {
      tagRejectFilter = trf;
    }
      
    public boolean accept(Dependency ob) {
      if (!(ob instanceof Dependency)) {
        return false;
      }
      Dependency d = (Dependency) ob;
      if ( ! (d.dependent() instanceof HasTag)) {
        return false;
      }
      String tag = ((HasTag) d.dependent()).tag();
      return tagRejectFilter.accept(tag);
    }

  } // end class DependencyRejectFilter

  // extra class guarantees correct lazy loading (Bloch p.194)
  private static class ComparatorHolder {

    private static class DependencyIdxComparator implements Comparator<Dependency> {

      public int compare(Dependency dep1, Dependency dep2) {
        AbstractMapLabel dep1lab = (AbstractMapLabel) dep1.dependent();
        AbstractMapLabel dep2lab = (AbstractMapLabel) dep2.dependent();
        int dep1idx = dep1lab.index();
        int dep2idx = dep2lab.index();
        return dep1idx - dep2idx;
      }

    }
    
    private static final Comparator dc = new DependencyIdxComparator();
  
  }

}
